#include "dfeed.h"
#include "global.h"
#include "md5.h"
enum {F_OPEN,F_MODE,F_SENDING,F_COMPRESS};
static int bkt_total[MAX_DBN];
void xpending_release(Que *q, Flink *l, int line);
#define pending_release(q,f) xpending_release(q,f,__LINE__);
int send_limit(int msec);

int mydiv(int n, int block, int *rem, int *quot);
static Feed feeds[MAX_FEED];
static int nfeeds;
static int nscan;
static int maxque;
static time_t cur_tick;
static int trysend;
static char *g_data;
static Que *outoforder=NULL;
char *trimit(char *s);

#define MAX_PER_PACKET 8000
static int udp_packet=MAX_PER_PACKET;
#define HEAD_LEN 28
/*
~00008761:0000000017:000958:
^
*/

void xpending_release(Que *q, Flink *l, int x)
{
	if (l->cur==q) {
		imsg("Thats odd, releasing que entry I was sending\n");
		l->cur = NULL;
	}
	que_free(q);
}

int feeds_free(void)
{
	Feed *f;
	int i=0;
	for (;i<nfeeds;i++) {
		f = &feeds[i];
		if (f->flog!=NULL) fclose(f->flog);
		f->flog = NULL;
		if (f->fscan!=NULL) fclose(f->fscan);
		f->fscan = NULL;
		links_free(f->links);
		f->links = NULL;
		free(f->site);
		free(f->to);
		free(f->bind);
	}
	nfeeds = 0;
	return TRUE;
}
void links_free(Flink *l)
{
	Flink *next;
	for (;l!=NULL;l=next) { 
		next = l->next;
		chan_release(l->chan);
		chan_close(l->chan);
		fque_free(l);
		if (l->last_data!=NULL) free(l->last_data); l->last_data = NULL;
		free(l);
	}
}
void fque_free(Flink *l)
{
	Que *fq,*next;
	for (fq = l->que; fq!=NULL; fq=next) {
		next = fq->next;	
		que_free(fq);
	}
	l->que = NULL;
}
void feeds_exit(void)
{
	feeds_free();
	if (g_data!=NULL) free(g_data);
	g_data = NULL;
}
int feeds_load(void)
{
	char type[BFSZ],bind[BFSZ],to[BFSZ],site[BFSZ];
	int nchan;
	char cmd[BFSZ],p1[BFSZ];
	FILE *f;
	char bf[BFSZ];
	char *s;
	int islog;
	int compress=FALSE;
	int udp_speed,udp_ttl;
	/* Read newsfeeds.conf and create feeds structures */

	trysend = atoi(ini_get(I_TRYSEND));

	maxque = atoi(ini_get(I_DFEED_NQUE));
	if (maxque==0) maxque = 15; 

	strcpy(type,"");
	strcpy(to,"");
	strcpy(bind,"");
	udp_speed  = udp_ttl = 0;
	islog = FALSE;
	nchan = 1;
	
	f = fopen(file_config("newsfeeds.conf"),"r");
	if (f==NULL) {emsg("Cannot read %s %s\n",file_config("newsfeeds.conf"),strerror(errno)); return FALSE;}
	
	feeds_free();
	for (;!feof(f);) {
		if (fgets(bf,BFSZ-1,f)==NULL) break;
		s = strtok(bf," \t\r\n"); if (s==NULL) continue;
		ncpy(cmd,s,100);
		s = strtok(NULL," \t\r\n"); 
		if (s==NULL) strcpy(p1,"NONE");
		else ncpy(p1,s,100);
		strlwr(cmd);
		if (strcmp(cmd,"site")==0) {
			if (strcmp(type,"dfeed")==0) feeds_load_site(site,to,bind,nchan,islog,udp_speed,udp_ttl,compress);
			strcpy(type,"");
			strcpy(to,"");
			strcpy(bind,"");
			nchan = 1;
			islog = FALSE;
			compress = FALSE;
			udp_speed  = udp_ttl = 0;
			strcpy(site,p1); continue;
		}
		if (strcmp(cmd,"to")==0) ncpy(to,p1,100);
		strlwr(p1);
		if (strcmp(cmd,"type")==0) ncpy(type,p1,100);
		if (strcmp(cmd,"udp_speed")==0) udp_speed = atoi(p1);
		if (strcmp(cmd,"udp_ttl")==0) udp_ttl = atoi(p1);
		if (strcmp(cmd,"udp_packet")==0) udp_packet = atoi(p1);
		if (strcmp(cmd,"bind")==0) ncpy(bind,p1,100);
		if (strcmp(cmd,"nchan")==0) nchan = atoi(p1);
		if (strcmp(cmd,"log")==0) {islog = TRUE;}
		if (strcmp(cmd,"compress")==0) {compress = TRUE;}
	}
	fclose(f);
	if (udp_packet<=0) udp_packet=MAX_PER_PACKET;
	if (udp_packet>MAX_PER_PACKET) udp_packet=MAX_PER_PACKET;
	if (udp_packet!=MAX_PER_PACKET) imsg("UDP_PACKET size is %d\n",udp_packet);
	if (strcmp(type,"dfeed")==0) feeds_load_site(site,to,bind,nchan,islog,udp_speed,udp_ttl,compress);

	return TRUE;
}


int feeds_load_site(char *site, char *to, char *bind, int nchan, int islog, int udp_speed, int udp_ttl, int trycompress)
{
	Feed *f = &feeds[nfeeds++];

	imsg("feeds_load site (%s) to (%s) bind (%s) nchan %d islog %d udp_speed %d\n",site,to,bind,nchan,islog,udp_speed);

	memset(f,0,sizeof(*f));

	f->site = strdup(site);
	if (strlen(to)>0) f->to = strdup(to);
	else f->to = strdup(site);
	f->bind = strdup(bind);
	f->nchan = nchan;
	f->time_retry = 0;
	f->len = 0;
	f->in = NULL;
	f->links = NULL;
	f->n = nfeeds-1;
	/* Clear stats */
	f->nbytes = 0;
	f->ndup = 0;
	f->fid = -1;
	f->nrej = 0;
	f->nok = 0;
	f->ndel = 0;
	f->islog = islog;
	f->flog = NULL;
	f->isudp = FALSE;
	f->udp_speed = udp_speed;
	f->udp_ttl = udp_ttl;
	f->trycompress = trycompress;
	if (f->udp_speed>0) f->isudp = TRUE;
	if (nfeeds>100) nfeeds = 100;

	imsg("Created Feed %p %s\n",f,f->to);
	return TRUE;
}
enum {L_NONE, L_CHECK, L_CHECKED, L_SEND,L_SENDING,L_IHAVE};
enum {I_READY,I_SENT};

void link_send_check(Flink *l, Que *q)
{
	char bf[BFSZ];
	if (q->state!=L_CHECK) return;
	if (l->ihave!=I_READY) return;
	if (l->state!=F_SENDING) return;
	l->lastdata = cur_tick;
	if (!l->stream) {
		dmsg("send ihave %s\n",q->mid);
		sprintf(bf,"IHAVE %s\r\n",q->mid);
		/*dmsg("link_send_check %s",bf);*/
		if (strlen(bf) + l->outlen + 10 <BFSZ) {
			strcpy(l->outbf + l->outlen,bf);
			l->outlen += strlen(bf);
			q->state = L_CHECKED;
		}
		l->ihave = I_SENT;
		chan_setsending(l->chan,TRUE);
		return;
	}
	sprintf(bf,"CHECK %.200s\r\n",q->mid);
	/*imsg("link_send_check %s",bf);*/
	if (strlen(bf) + l->outlen + 10 <BFSZ) {
		strcpy(l->outbf + l->outlen,bf);
		l->outlen += strlen(bf);
		q->state = L_CHECKED;
	}
	chan_setsending(l->chan,TRUE);
}
int link_send_checks(Chan *c, Flink *l)
{
	Que *fq;
	for (fq = l->que; fq!=NULL; fq=fq->next) {
		if (fq->state==L_CHECK) link_send_check(l,fq);
	}
	if (l->outlen>0) return TRUE;
	return FALSE;
}
int feeds_write_next(Chan *c, Flink *l)
{
	Que *fq;
	int sentone = FALSE;

	if (l->cur==NULL) {
		link_que_fill(l);
		for (fq=l->que;fq!=NULL;fq=fq->next) {
			if (fq->state==L_SEND) {
				l->cur = fq;
				l->sendtake = 0;
				fq->state = L_SENDING;
				dmsg("state=sending %s\n",fq->mid);
				l->sendat = 0;
				chan_setsending(c,TRUE);
				return TRUE;
			}
			if (link_send_checks(c,l)) {
				chan_setsending(c,TRUE);
				sentone = TRUE;
				return TRUE;
			}


		}
		if (sentone) return TRUE;
		if ((l->outlen==0) && (strlen(l->rest)==0)) {
			chan_setsending(c,FALSE);
			return FALSE;
		}
	}
	chan_setsending(c,TRUE);
	return TRUE;
}
char *que_get_link(Flink *l, Que *q, int from, int *got, int *total, Feed *feed)
{
	int len;
	char *data;
	len = q->len - HEAD_LEN - 1;
	if (l->last_data!=NULL) if (strcmp(l->last_mid,q->mid)==0) {
		data = l->last_data;
		*total = len;
		*got = len - from; /* r - from;*/
		dmsg("link: found  (%s) len %d from %d tot %d\n",l->last_mid,*got,from,len);
		return data + from;
	}
	ncpy(l->last_mid,"",BFSZ-1);
	
	if (l->last_len==0) {
		l->last_len = atoi(ini_get(I_ITEM_MAX));
		if (l->last_len<1000000) l->last_len = 1000000;
		l->last_len += 10000;
		l->last_data = malloc(l->last_len);
	}
	if (len+1000>l->last_len) {
		emsg("GETFAIL Item is too large to send %d>%d, increase item_max in dnews.conf\n",len+1000,l->last_len);
		return NULL;
	}
	data  = que_get(q,from,got,total,feed);
	if (data!=NULL) {
		ncpy(l->last_mid,q->mid,BFSZ-1);
		if (*got+100>l->last_len) {
			emsg("GETFAIL Item2 is too large to send %d>%d, increase item_max in dnews.conf\n",len+1000,l->last_len);
			return NULL;
		}
		memcpy(l->last_data,data,*got+4);
		dmsg("link: saving (%s) len %d\n",l->last_mid,*got);
	}

	return data;

}
static char nulmsg[BFSZ];
static int nnul;
void feeds_write(Chan *c, void *p1)
{
	Flink *l=(Flink *) p1;
	char *data;
	int i;
	char bf[BFSZ];
	int got,total;
	Que *q;
	int rest;
	int r;
	int totsent = 0;
next_one:
	if (l->state != F_SENDING) return;
	imsg_feed(l->feed,l->id);
	if (strlen(l->rest)>0) {
		r = chan_write(l->chan,l->rest,strlen(l->rest));
		if (r<0) {
			imsg("Chan_write failed\n");
			goto ret;
		}
		if (r<(int)strlen(l->rest)) {
			memmove(l->rest,l->rest + r,strlen(l->rest+r)+1);
			goto ret;
		}	
		strcpy(l->rest,"");
	}

	l->lastdata = cur_tick;
	if (!feeds_write_next(c,l)) {
		imsg("Nothing to write to this channel just now\n");
		goto ret;
	}
	if (l->cur==NULL) {
  		if (l->outlen==0) {
			goto ret;
		}
		if (!l->stream) { chan_setdelay(l->chan,FALSE); }
		r = chan_write(l->chan,l->outbf,l->outlen);
		dmsg("feeds_write_outbf %d/%d bytes\n",r,l->outlen);
		if (r<0) { emsg("Channel closed while writing data\n"); goto ret; }
		if (r < l->outlen) {
			memmove(l->outbf,l->outbf+r,l->outlen-r+1);
			l->outlen -= r;
			goto ret;
		}
		l->outlen = 0;
		goto ret;
	}

	if (!chan_isopen(l->chan)) goto ret;
	data = que_get_link(l,l->cur,l->sendat,&got,&total,l->feed);
	if (data==NULL) {
isnull:	
		q = link_que_remove(l,l->cur->mid); 
		l->cur = NULL;	
		if (q!=NULL) pending_release(q,l);  /* Data is not needed by this channel */
		l->sendtake = 0;
		l->feed->ndel++;
		l->ihave = I_READY;
		if (!l->stream) {
			r = chan_write(l->chan,"\r\n.\r\n",5);
			link_sendmore(l);
			goto ret;
		}
		feeds_write_next(c,l);		
		goto ret;
	}

	for (i=0; i<got; i++) {
		if (data[i]==0) {
			nnul++;
			sprintf(nulmsg,"%s Null at %d, sendat=%d, got=%d total=%d, %s, bkt=%d offset=%d"
				,lib_showtime()
				,i,l->sendat,got,total
				,l->cur->mid,l->cur->bkt,l->cur->offset);
			emsg("%s\n",nulmsg);
			goto isnull;
		}
	}
	if (nnul>0) {
		imsg("Nulls found %d %s\n",nnul,nulmsg);
	}
	/* Get a chunk of this message, with big ones it might not be cached
	if 'x' mb is already cached */

	dmsg("Sending %s %d bytes %d\n",l->cur->mid,got,l->sendtake);
	if ((l->sendtake!= -1) && l->stream) {
		que_show(l);
		sprintf(bf,"TAKETHIS %s\r\n",l->cur->mid);
		rest  =  strlen(bf+l->sendtake);
		r = chan_write(l->chan,bf+l->sendtake,rest);
		if (r<0) {
			emsg("takethis write failed %s\n",tcperror());
			link_close(l,tcperror());
			goto ret;
		}
		if (r<rest) {
			l->sendtake += r;
			goto ret;
		}
		l->sendtake = -1;
	}
	chan_setinitem(TRUE);
	r = chan_write(l->chan,data,got);

//	if (got>10000) 	r = chan_write(l->chan,data,10000);
//	else r = chan_write(l->chan,data,got);

	dmsg("wrote: %d/%d total=%d \n",r,got,total);
	chan_setinitem(FALSE);
	totsent += got; /* this is really correct don't change to 'r' */
	if (r<0) { emsg("Channel closed while writing data\n"); goto ret; }
	l->sendat += r;
	l->feed->nbytes += r; /* Bytes sent */
	if (l->sendat >= total) {
		sprintf(bf,".\r\n",3);
		r = chan_write(l->chan,bf,strlen(bf));		
		if (r<0) {
			goto ret;	
		}
		if (r<(int) strlen(bf)) {
			strcpy(l->rest,bf+r);
		}
		l->cur = NULL;
		l->sendtake = 0;
		feeds_write_next(c,l);
		if (l->cur!=NULL) {
			if (totsent<trysend) {
				imsg("And next one %d/%d sent now\n",totsent,trysend);
				goto next_one;
			}
		}
	}
ret:;
	link_ifempty(l);
	imsg_feed(NULL,0);

}
void link_ifempty(Flink *l)
{
	if (l->que==NULL) link_sendmore(l);
}

void feeds_idle(void)
{
	/* Open channels to outgoing feeds */
	static int last;
	int fi;
	Flink *l;
	time_t t = time(NULL);

	cur_tick = t;
	if (t-last<10) return;

	last = t;
	
	imsg_flush();
	feeds_idle_open();

	stats_start(S_FEEDS_SCAN);
	for (fi=0; fi<nfeeds; fi++) {
		imsg_feed(&feeds[fi],0);
		feeds_scan(&feeds[fi]);
		for (l=feeds[fi].links;l!=NULL;l=l->next) {
			link_que_fill(l);
			link_que_trim(l);
		}
		imsg_feed(NULL,0);
	}
	stats_end(S_FEEDS_SCAN);	
	if (lib_exists(file_work("dfeed.exit"))) {
		dfeed_exit();
	}
}
void udp_idle_one(Feed *f);
void udp_idle(void)
{
	int fi;
	for (fi=0; fi<nfeeds; fi++) {
		if (!feeds[fi].isudp) continue; /* added by chrisp */
/*		imsg("udp_idle %p %s [%d]\n",&feeds[fi],feeds[fi].to,fi); imsg_flush();*/

		imsg_feed(&feeds[fi],0);
		udp_idle_one(&feeds[fi]);
		imsg_feed(NULL,0);
	}
}
static int udp_delay;
void udp_send(Flink *l, char *data, int len);
void udp_idle_one(Feed *f)
{
	Que *q;
	char *data;
	Flink *l = f->links;
	int got,total;
	int delay;

	if (l==NULL) return;
	if (!f->isudp) return;
	if (!chan_isopen(l->chan)) return;
	delay =  (udp_packet * 1000) / f->udp_speed ;
	if (delay<=0) delay = 1;
nextone:
	q = link_next(l);
	if (q==NULL) {return;}
	data = que_get(q,0,&got,&total,l->feed);
	if (data==NULL) {
		que_free(q);
		goto nextone;
	}
/*	tcp_nextwait(delay/2); */
	tcp_nextwait(delay/2); 
	udp_delay = delay; 	

	udp_send(l,data,got);
	que_free(q);
}
char *undot_stuff(char *in, int *len)
{
	static int datalen;
	static char *outbf;
	char *out;
	char *xin=in;
	int n = *len;


	if (outbf==NULL) {
		datalen = atoi(ini_get(I_ITEM_MAX));
		if (datalen<1000000) datalen = 1000000;
		datalen += 10000;
		outbf = malloc(datalen);
	}
	for (out=outbf;in<(xin+n); ) {
		if (*in=='\r') if (in[1]=='\n') in++; /* convert crlf to lf */
		*out++ = *in;
		if (*in=='\n') if (in[1]=='.') in++;
		in++;
		if (out-outbf>=(datalen-2)) return NULL;
	}
	*out = 0;
	*len = out-outbf;
	return outbf;
}


int udp_send_small(Flink *l, char *data, int len);
int udp_send_large(Flink *l, char *data, int len);
void udp_send(Flink *l, char *data, int len)
{
	
	data = undot_stuff(data,&len);
	if (data==NULL) {
		imsg("udp: Message too big after dot stuffing\n");
		return;
	}
/*
	f = fopen("udp.stuff","wb");
	fwrite(data,1,len,f);
	fclose(f);
*/
	if (len > udp_packet) udp_send_large(l,data,len);
	else udp_send_small(l,data,len);

	l->feed->nok++;
	ncpy(l->feed->reason,"Sent Message Ok",100);
	l->feed->nbytes+= len;

}
/*
	1 byte article type = 0
	2 bytes data size = N (rem, quot)
	N bytes article
*/
static char xbuffer[MAX_PER_PACKET + 3];
int udp_send_small(Flink *l, char *data, int len)
{
	int rem,quot;
	/* Fill in rest of packet */
	xbuffer[0] = 0;

	mydiv(len, 256,&rem,&quot);
	xbuffer[1] = rem;
	xbuffer[2] = quot;
	memcpy(xbuffer+3,data,len);
	tcp_setblockraw(l->chan->sk,TRUE);
	errno = 0;
	send_limit(udp_delay);
	if (chan_write(l->chan,xbuffer,len+3)<=0) {
		imsg("udp_send_small failed (%d bytes)\n",len+3);
	} 
/*imsg("send_small packet %d bytes (%.16s)\n",len+3,xbuffer+3);*/
	return TRUE;
}
/*
	1 byte article type = 2
	2 bytes data size = N (rem, quot)
	16 bytes hash of article
	1 byte part number of this part
	1 byte total number of parts
	N bytes partial article
*/
int fill_in_hash(char *hash_buffer, char *data, int len)
{
	MD5_CTX md5_ctx;
	int i;
	unsigned char digest[16];
	unsigned char hash_text[33];


	MD5Init(&md5_ctx);
	strcpy(hash_text, "");
	MD5Update(&md5_ctx, data, len);

	MD5Final(digest, &md5_ctx);
	for (i = 0; i < 16; i++) {
		sprintf(hash_text + (i * 2), "%02x", digest[i]);
	}
	memcpy(hash_buffer, digest, 16);
	return TRUE;
}
int mydiv(int n, int block, int *rem, int *quot)
{
	*quot = n / block;
	*rem = n % block;
	return TRUE;
}
int udp_send_large(Flink *l, char *data, int len)
{
	int rem,quot;
	int parts_total,parts_count,upto;
	int xchunk;

	xbuffer[0] = 2;
	mydiv(len, 256,&rem,&quot);
	xbuffer[1] = rem;
	xbuffer[2] = quot;

	fill_in_hash(xbuffer + 3, data,len);
	parts_total = (len-1)  / (udp_packet - 21);
	if (parts_total>255) {
		emsg("Could not send article %d bytes via udp\n",len);
		return FALSE;
	}
	xbuffer[20] = parts_total + 1;

	upto = 0;
	for (parts_count = 0; parts_count <= parts_total; parts_count++) {
		xchunk = udp_packet - 21;
		if (xchunk+upto>len) xchunk = len - upto;
		if (xchunk==0) break;


		mydiv(xchunk, 256,&rem,&quot);

		xbuffer[1] = rem;
		xbuffer[2] = quot;
		memcpy(xbuffer+21,data+upto,xchunk);
		upto += xchunk;
		xbuffer[19] = parts_count + 1;
/* imsg("udpsending %d/%d\n",parts_count,parts_total);*/
		send_limit(udp_delay);
		tcp_setblockraw(l->chan->sk,TRUE);
		if (chan_write(l->chan,xbuffer,xchunk+21)<=0) {
			imsg("udp_send_large failed %d bytes\n",xchunk+21);
			break;
		}
/*imsg("send_large packet %d bytes (%.16s)\n",xchunk+21,xbuffer+3);*/
	}
	return TRUE;

}
int send_limit(int msec)
{
	static int last;
	int now;
	int want;
	int gap;

/*lib_msleep(msec);
//return TRUE;*/
	if (last==0) last = lib_msec();

	now = lib_msec();
	gap = now - last;

	last = now;
	if (gap<msec) {
		want = msec - gap;
		if (want>2000) want = 2000;
		if (want<0) return TRUE;
		lib_msleep(want);
	}
	return TRUE;
}
/*
Wed Jan 24 08:25:10 239.2.0.31:1119:0 send_large (4239)
Wed Jan 24 08:25:10 239.2.0.31:1119:0 Send packet chunk=4075 len=4239 upto=0
Wed Jan 24 08:25:10 239.2.0.31:1119:0 div1Send packet chunk=4075 len=4239 upto=0
Wed Jan 24 08:25:10 239.2.0.31:1119:0 div2Send packet chunk=235 len=4239 upto=0 n=15

  Wed Jan 24 08:25:10 239.2.0.31:1119:0 aSend packet chunk=0 len=4239 upto=0
Wed Jan 24 08:25:10 239.2.0.31:1119:0 bSend packet chunk=0 len=4239 upto=0
Wed Jan 24 08:25:10 239.2.0.31:1119:0 cSend packet chunk=0 len=4239 upto=0
Wed Jan 24 08:25:10 239.2.0.31:1119:0 dSend packet chunk=0 len=4239 upto=0
Wed Jan 24 08:25:10 239.2.0.31:1119:0 About to write chunk 22 bytes

Wed Jan 24 08:25:10 239.2.0.31:1119:0 Send packet chunk=0 len=4239 upto=0
*/

void link_close(Flink *l, char *reason)
{
	Que *q;
	time_t t;
	struct tm *xtm;
	char bf[BFSZ];
	t = time(NULL);
	xtm = localtime(&t);

	for (q=l->que;q!=NULL;q=q->next) {
		imsg("CloseEntry? %s state=%d %d:db_%d_%d.itm at=%d len=%d \n",
			q->mid,q->state,q->spool,q->bkt,q->pile,q->offset,q->len);
		q->ntries+= 100;
		feed_reque(q,l->feed);
	}
	

	imsg("%02d:%02d:%02d Closing link %.80s\n",xtm->tm_hour,xtm->tm_min,xtm->tm_sec,trimit(reason));
	sprintf(bf,"%02d:%02d:%02d %s",xtm->tm_hour,xtm->tm_min,xtm->tm_sec,trimit(reason));
	fque_free(l);
	if (chan_isopen(l->chan)) chan_close(l->chan);
	ncpy(l->feed->reason,bf,100);
	ncpy(l->feed->problem,bf,100);
	l->inlen = 0;
	l->outlen = 0;
	l->ihave = I_READY;
	l->feed->time_retry = time(NULL) + 10; /* Change to 60*/
	l->nretry = 0;
	l->sendat = 0;
	l->state = F_OPEN;
	l->cur = NULL;
	l->sendtake = 0;
	l->nque = 0;
	strcpy(l->in,"");
	strcpy(l->outbf,"");
	strcpy(l->rest,"");
	l->stream = 0;
	l->wait = 0;
	l->lastoffset = 0;
	l->lastdata = time(NULL)+10;
}


Flink *link_new(void)
{
	Flink *l;
	l = malloc(sizeof(Flink));
	memset(l,0,sizeof(Flink));
	return l;
}
int feeds_handles(void)
{
	/* Open channels to outgoing feeds */
	int i;
	int n=0;
	Feed *f;
	Flink *l;
	for (i=0; i<nfeeds; i++) {
		f = &feeds[i];
		for (l=f->links;l!=NULL;l=l->next) {
			if (l->chan==NULL) continue;
			if (chan_isopen(l->chan)) n++;
		}
	}
	return n;
}
void feeds_idle_open(void)
{
	/* Open channels to outgoing feeds */
	int i;
	char bf[BFSZ];
	Feed *f;
	char logname[BFSZ];
	int nlink=0;
	Flink *l;
	int t=time(NULL);
	for (i=0; i<nfeeds; i++) {
		f = &feeds[i];
		imsg_feed(f,0);
		if (f->islog) {
			ncpy(logname,file_feedlog(f->site),200);
			if (f->flog==NULL) {
				f->flog = fopen(logname,"ab");
				emsg("Open log file %s %s\n",logname,strerror(errno));
				dmsg("\n");
				dmsg("\n");
				dmsg("------ DFEED %s Starting/reloading -------\n",DFEED_VERSION);
			}
			if (f->flog!=NULL) {
				if (file_size(f->flog)>2000000) {
					fclose(f->flog);
					f->flog = fopen(logname,"wb");
				}
				if (f->flog!=NULL) fflush(f->flog);
			}
		}
		if (f->time_retry>t) continue; 
		if (f->links==NULL) {
		 for (i=0; i<f->nchan; i++) {
			l = link_new();
			l->feed = f;
			l->next = f->links;
			l->id = i;
			f->links = l;
		 }
		}

		for (l=f->links;l!=NULL;l=l->next) {
			imsg_feed(l->feed,l->id);
			nlink++;
			if (l->chan==NULL) l->chan = chan_new();
			if (chan_isopen(l->chan)) continue;
			f->time_retry = t + 10;
			strcpy(l->feed->reason,"Trying to open link");
			if (f->isudp) {
				chan_open_udp(l->chan,f->to,f->bind);
				if (f->udp_ttl>0) if (l->chan->sk!=INVALID_SOCKET) {
					imsg("Setting ttl to %d\n",f->udp_ttl);
					tcp_multicast_ttl(l->chan->sk, f->udp_ttl);
				}
				chan_setread(l->chan,feeds_read_udp,l);
				chan_setwrite(l->chan,feeds_write_udp,l);
			} else {
				chan_open(l->chan,f->to,f->bind);
				chan_setread(l->chan,feeds_read,l);
				chan_setwrite(l->chan,feeds_write,l);
			}
			l->state = F_OPEN;
			l->stream = FALSE;
			l->lastdata = cur_tick;
		}

		for (l=f->links;l!=NULL;l=l->next) {

			imsg_feed(l->feed,l->id);
/*			if (l->state!=F_SENDING) l->lastdata = cur_tick;*/
			if ((cur_tick - l->lastdata)>120) {
				imsg("Closing idle channel feed %s %d qued\n",f->site,l->nque);
				sprintf(bf,"Idle, %d queued",l->nque);
				link_close(l,bf);
			}
		}

	}
	imsg_feed(NULL,0);
}
/* 

  Read data from socket 

*/
void feeds_write_udp(Chan *c, void *p1)
{
	imsg("feeds_write_udp\n");
	lib_msleep(5);

}
void feeds_read_udp(Chan *c, void *p)
{
	imsg("feeds_read_udp\n");
	lib_msleep(5);
}
void feeds_read(Chan *c, void *p)
{
	int r;
	char bf[BFSZ];
	Flink *l = (Flink *) p;
	Feed *f = l->feed;
	int len;
	int rest;
	char *s;


	imsg_feed(l->feed,l->id);
	
	len = BFSZ-100-l->inlen;
	if (len<0) {
		emsg("Cannot be true %d\n",l->inlen);
		abort();
	}
	r = chan_read(c,l->in+l->inlen,len);
	if (r<=0) {
		if (!chan_isopen(c)) {
			emsg("Channel closed %s\n",tcperror());
			sprintf(bf,"Channel closed %s",tcperror());
			link_close(l,bf);
		}
		goto ret;
	}
	l->lastdata = cur_tick;

	l->inlen += r;
again:
	s = mystrstr(l->in,"\r\n");
	if (s==NULL) {
		if (l->inlen>300) {
			emsg("Thats odd, connection sent 300 char with no end of line\n");
			link_close(l,"Long line");
		}
		goto ret;
	}
	*s = 0;

	rest = l->inlen - ((s+2)-l->in);
	feeds_read_line(c,f,l,l->in);
	if (!chan_isopen(c)) return;
	memmove(l->in,s+2,rest+1);
	l->inlen = rest;
	if (l->inlen>0) goto again;

ret:;
	imsg_feed(NULL,0);
}
char *trimit(char *s)
{
	static char bf[BFSZ];
	ncpy(bf,s,100);
	return bf;
}
void feeds_read_line(Chan *c, Feed *f, Flink *l, char *data)
{
	int code = atoi(data);
	dmsg("Read %s\n",trimit(data));
	switch(l->state) {
 	  case F_OPEN: /* 200, 201, 480, 500... */
		  if ((code>299) || (code<200)) {
			imsg("Open to %s failed %s\n",f->to,trimit(data));
			link_close(l,data);
		  }
		chan_write_s(c,"MODE STREAM\r\n");
		strcpy(l->feed->reason,"Got Welcome");
		l->state = F_MODE;
		break;
	  case F_MODE:
		/* 500 Syntax error*/
		if (code==500) l->stream = FALSE;
		/* 203 StreamOK */
		if (code==203) l->stream = TRUE;
		/* 
			if (code==203) l->stream = FALSE;
			imsg("Debugging, sending in non streaming mode\n");
		*/
		if (l->stream) imsg("Streaming mode set\n");
		else imsg("IHAVE mode set\n");
		if (l->feed->trycompress) {
#ifdef DOCOMPRESS
			imsg("Sending mode compress to see if the other end can receive a compressed feed\n");
			l->state = F_COMPRESS;
			chan_write_s(c,"mode compress\r\n");
			break;
#else
			emsg("Sorry, this build/platform of dfeed does not include support for compressed feeds\n");
#endif
		}
		l->state = F_SENDING;
		strcpy(l->feed->reason,"Got Stream Response - streaming ok");
		if (!l->stream) strcpy(l->feed->reason,"Got Stream Response - cannot stream");
		link_send(l);
		feeds_write_next(c,l);
		break;
	  case F_COMPRESS:
		if (code==207) {
			imsg("Yes, destination supports compressed feeds\n");
			c->iscompress = TRUE;
#ifdef DOCOMPRESS
			memset(&c->z_str,0,sizeof(z_stream));
			deflateInit(&c->z_str,Z_DEFAULT_COMPRESSION);
#endif
			strcpy(l->feed->reason,"Mode stream response, Compressed feed ok");
		} else {
			imsg("Sorry, destination does not support compressed feeds\n");
			strcpy(l->feed->reason,"Mode stream response, Compressed feed NOT SUPPORTED");
		}
		l->state = F_SENDING;
		link_send(l);
		feeds_write_next(c,l);
		break;
	  case F_SENDING:
		feeds_read_sending(c,f,l,data);
		break;
	}
}

/*

*/
void link_send(Flink *l)
{
	link_que_fill(l);


}
void feeds_read_sending(Chan *c, Feed *f, Flink *l, char *status)
{
	int code = atoi(status);

	ncpy(l->feed->reason,trimit(status),60);
	
	switch (code) {
	/* Stream Check*/
	  case 438: /* <mid> article not wanted */
		link_dup(l,get_word(status,1));
		break;
	  case 238: /* <mid> yes, send article */
		link_takethis(l,get_word(status,1));
		break;
	/* Stream takethis */
	  case 431: /* <mid> try later  */
		link_later(l,get_word(status,1));
		break;
	  case 439: /* <mid> rejected  */
		link_rej(l,get_word(status,1));
		break;
	  case 239: /* <mid> transferred ok */
		link_ok(l,get_word(status,1));
		break;
	/* Ihave check */
	  case 435: /* Not wanted, duplicate */
		link_ihave_dup(l);
		break;
	  case 335: /* Yes, send it */
		link_ihave_send(l);
		break;
	/* Ihave sent data */
	  case 437: /* Rejected  */
		link_ihave_rej(l);
		break;
	  case 235: /* Transferred ok */
		link_ihave_ok(l);
		break;
	  case 436: /* Paused/full, try later  */
		imsg("Temporary problem, will try later (%s), closing link\n",trimit(status));
		link_later(l,NULL);
		goto paused;
	  default:  /* Misc errors */
		    /* 	4xx close connection and try later	*/
		    /*	  480 Permission denied		*/
		imsg("Bad Status (%s), closing link\n",trimit(status));
paused:		link_close(l,trimit(status));
		f->time_retry = time(NULL) + 60;
		break;
	}

}
void log_fault(Flink *l, Que *q, char *mid);

Que *link_que_remove(Flink *l, char *mid)
{
	Que *q;
	Que *fq;
	int done=FALSE;
	Que **last=&l->que;
	outoforder = NULL;

	for (fq = l->que; fq!=NULL; fq=fq->next) {
		if (mid==NULL) {
			if (fq->state==L_CHECKED) goto thisone;
			if (fq->state==L_SENDING) goto thisone;
			goto cont;
		}
		if (strcmpnc(mid,fq->mid)==0) {
			if (done) {
				if (fq->state==L_SENDING) log_fault(l,fq,mid);
			}
thisone:		*last = fq->next;
			q = fq;
			l->nque--;			
			return q;
		}
		if (fq->state==L_SENDING) if (!done) {
			done = TRUE;
		}
cont:		last = &fq->next;
	}
	if (mid==NULL) mid = "ihave";
	emsg("FATAL ERROR, could not find that entry in the que (%s) \n",mid);
	return 	NULL;
}
void log_fault_one(FILE *F, Que *q,Feed *feed);
void log_fault(Flink *l, Que *q, char *mid)
{
	/* A fault seems to have occured, so lets write
	out all the que entries up to the right mid to a file
	
	*/
	char logname[BFSZ];
	static time_t lasttime;
	Que *fq;
	FILE *f;
	Feed *fd = l->feed;
	static int done;
	if (done) return;
	done = TRUE;

	if (time(NULL)-lasttime<200) return;
	que_show(l);

	lasttime = time(NULL);
	f = fopen(file_work("dfeed.ord"),"wb");
	if (f==NULL) return;
	for (fq=l->que; fq!=NULL; fq=fq->next) {
		fprintf(f,"Entry in Queue %s\r\n",fq->mid);
		log_fault_one(f,fq,l->feed);
		fprintf(f,"}End of entry\r\n\r\n");
		if (strcmp(fq->mid,mid)==0) break;
	}
	fclose(f);

	if (fd->flog!=NULL) fclose(fd->flog);
	fd->flog = NULL;
	ncpy(logname,file_feedlog(fd->site),200);
	remove(file_work("dfeed.copylog"));
	imsg("Response in wrong order, trying to copy log file %s to %s\n",logname,file_work("dfeed.copylog"));
	if (rename(logname,file_work("dfeed.copylog"))!=0) {
		imsg("Rename failed %s\n",strerror(errno));	
	}
}
void log_fault_one(FILE *f, Que *q,Feed *feed)
{
	char *body;
	int got=0;
	int total=0;
	body = que_get(q, 0, &got,&total,feed);
	if (body==NULL) return;
	fwrite(body,1,total,f);
}
void link_rej(Flink *l, char *mid)
{
	Que *q;
	q = link_que_remove(l,mid); if (q==NULL) return ;
	l->feed->nrej++;

	pending_release(q,l);	/* Record that item is no longer needed for this feed */
	link_sendmore(l);
}
void link_dup(Flink *l, char *mid)
{
	Que *q;
	q = link_que_remove(l,mid); if (q==NULL) return ;
	l->feed->ndup++;

	pending_release(q,l);	/* Record that item is no longer needed for this feed */
	link_sendmore(l);
}
Que *link_que_find(Flink *l,char *mid)
{
	Que  *q;
	for (q=l->que;q!=NULL;q=q->next) {
		if (mid==NULL) return q;
		if (strcmp(q->mid,mid)==0) return q;
	}
	return NULL;
}
Que *link_que_find_state(Flink *l,int state)
{
	Que  *q;
	for (q=l->que;q!=NULL;q=q->next) {
		if (q->state==state) return q;
	}
	return NULL;
}
void link_takethis(Flink *l, char *mid)
{
	Que *q;
	q = link_que_find(l,mid); if (q==NULL) return ;
	q->state = L_SEND; /* Send item when possible */
	chan_setsending(l->chan,TRUE); /* */ 
}
void link_later(Flink *l, char *mid)
{
	Que *q;
	q = link_que_remove(l,mid); if (q==NULL) return ;
	l->feed->nlater++;
	l->nretry++;
	feed_reque(q,l->feed); /* append q entry back in bucket queue file */
	pending_release(q,l);	/* Record that item is no longer needed for this feed */
	link_sendmore(l);
	if (l->nretry>100) {
		emsg("Trying that link later as 100 try later errors occured\n");
		link_close(l,"Told to try later");
		l->feed->time_retry = time(NULL) + 300;
		l->nretry = 0;
	}
}
void link_ok(Flink *l, char *mid)
{
	Que *q;
	q = link_que_remove(l,mid); if (q==NULL) return;
	l->nretry = 0;
	l->feed->nok++;

	if ((l->feed->nok % 300) == 2) {
		feed_write_upto(l->feed,mid);
	}
	ncpy(l->feed->reason,"Sent Message Ok",100);

	l->feed->nlater = 0;
	pending_release(q,l);	/* Record that item is no longer needed for this feed */
	link_sendmore(l);
}
int link_isfull(Flink *l)
{
	if (l->nque>=maxque) return TRUE;
	return FALSE;
}
void link_sendmore(Flink *l)
{
	dmsg("link_sendmore, %d currently ready to send\n",l->nque);
	if (!l->stream) {
		link_send_checks(l->chan,l);
	}
	if (link_isfull(l)) return;
	if (l->nque> maxque/2) return ; /* So we always get at least 5 */
	/*if (l->wait>time(NULL)) return;*/
	l->wait = time(NULL) + 1;
	link_que_fill(l);
}

void link_ihave_rej(Flink *l)
{
	Que *q;

	q = link_que_remove(l,NULL); if (q==NULL) return;
	l->feed->nrej++;
	l->ihave = I_READY;
	pending_release(q,l);	/* Record that item is no longer needed for this feed */
	link_sendmore(l);

}
void link_ihave_dup(Flink *l)
{
	Que *q;
	q = link_que_remove(l,NULL); if (q==NULL) return;
	l->feed->ndup++;
	l->ihave = I_READY;
	pending_release(q,l);	/* Record that item is no longer needed for this feed */
	link_sendmore(l);
}
void link_ihave_send(Flink *l)
{
	Que *q;
	q = link_que_find_state(l,L_CHECKED); if (q==NULL) return;
	q->state = L_SEND; /* Send item when possible */
	
	chan_setsending(l->chan,TRUE); /* */ 
}

void link_ihave_ok(Flink *l)
{
	Que *q;
	q = link_que_remove(l,NULL); if (q==NULL) return;
	l->feed->nok++;
	l->ihave = I_READY;
	pending_release(q,l);	/* Record that item is no longer needed for this feed */
	link_sendmore(l);
}

	/* 
	  
		Delete queues for deleted buckets somehow.

		If dbn file changed, then read and figure out
		which buckets no longer exist, and delete
		the queues.
	*/

	/*
Format of bucket feed files ? 
  feed.name <336@netwin.co.nztest.12> test.12:4:5:17600:8800:0
  sitename  mid group:item:bucket:offset:len:spool 
*/

void feeds_scan(Feed *feed)
{
/*
	scan for waiting .send files

	If found rename and start processing.
	
*/
	char sendtmp[BFSZ];
	char sendfile[BFSZ];

	if (feed->fscan!=NULL) return;
	ncpy(sendfile,file_send(feed->site),200);
	ncpy(sendtmp,file_sendtmp(feed->site),200);
	if (lib_exists(sendtmp)) {
again:		imsg("Opening send file %s\n",sendtmp);
		feed->fscan = fopen(sendtmp,"r");
		if (feed->fscan==NULL) {
			emsg("Cannot read (%s) (%s) \n",sendtmp,strerror(errno));
			return;
		} else {
			feeds_scan_upto(feed);
		}
		return;
	}
	imsg("Searching for (%s)\n",sendfile);
	if (!lib_exists(sendfile)) return;
	imsg("Found it, renaming it\n");
	if (rename(sendfile,sendtmp)!=0) { imsg("Rename failed (%s) (%s) (%s)\n",sendfile,sendtmp,strerror(errno)); return; }
	goto again;
}
void feeds_scan_upto(Feed *feed)
{
	int len;
	char bf[BFSZ];
	FILE *f = feed->fscan;
	char *upto = feed_read_upto(feed);
	if (upto==NULL) return;
	len = strlen(upto);

	emsg("Scanning for upto restart mid %.80s\n",upto);
	if (f==NULL) return;
	for (;!feof(f);) {
		if (fgets(bf,BFSZ-10,f)==NULL) break;
		if (strncmp(bf,upto,len)==0) {
			emsg("Found restart point %.80s\n",upto);
			return;
		}
	}
	fseek(f,0,SEEK_SET);
	remove(file_sendupto(feed->site));
	emsg("Couldn't find restart point %.80s\n",upto);

}
int send_read(FILE *f, char *mid, int *bkt, int *offset, int *len, int *spool, int *pile, int *ntries)
{
	char xbf[BFSZ];
	char *s;
	*ntries = 0;
skipline:
	if (feof(f)) {
		imsg("send_read_eof\n");
		return FALSE;
	}
	if (fgets(xbf,200,f)==NULL) {
		imsg("send_read, read null, probably eof\n");
		return FALSE;
	}
	/*mid*/ s = strtok(xbf," "); if (s==NULL) return FALSE;
	ncpy(mid,s,100);
	if (strlen(s)>100) return FALSE;
	/*grp*/ s = strtok(NULL,":"); if (s==NULL) return FALSE;
	/*n  */ s = strtok(NULL,":"); if (s==NULL) return FALSE;
	/*bkt*/ s = strtok(NULL,":"); if (s==NULL) return FALSE;
	*bkt = atoi(s);
	/*offset*/ s = strtok(NULL,":"); if (s==NULL) return FALSE;
	*offset = atoi(s);
	/*len*/ s = strtok(NULL,":"); if (s==NULL) goto skipline;
	*len = atoi(s);
	/*spool*/ s = strtok(NULL,":"); if (s==NULL) goto skipline;
	*spool = atoi(s);
	if (*spool < 0) return FALSE;
	if (*spool > 30) return FALSE;
	/*pile*/ s = strtok(NULL,": "); if (s==NULL) goto skipline;
	*pile = atoi(s);
	/* ntries */ s = strtok(NULL," "); 
	if (s!=NULL) {	
		if (strncmp(s,"tries=",6)==0) {
			*ntries = atoi(s+6);
		}
	}

	nscan++;
	return TRUE;
}


void feed_reque(Que *q, Feed *feed)
{
	FILE *f;
	if (q->ntries>300) {
		imsg("Dropping %.100s %.100s:%d:%d:%d:%d:%d:%d tries=%d\n",q->mid,"none",1,q->bkt,q->offset,q->len,q->spool,q->pile,q->ntries);
		return;	
	}
	f = fopen(file_send(feed->site),"a");
	if (f==NULL) return;
	fprintf(f,"%s %s:%d:%d:%d:%d:%d:%d tries=%d\n",q->mid,"none",1,q->bkt,q->offset,q->len,q->spool,q->pile,++(q->ntries));
	fclose(f);
}



/*	
	
	Que *old = link_que_find(l,q->mid);
	if (old!=NULL) {
		imsg("Ignoring duplicate entry in feed file %s\n",q->mid);
		return;
	}
	q->next = l->que;
	l->que = q;
	q->state = L_CHECK;
	link_send_check(l,q);
	l->nque++;
*/
void link_add_msg(Flink *l,Que *qnew)
{
	Que  *q,*last=NULL;
	for (q=l->que;q!=NULL;q=q->next) {
		last = q;
		if (strcmp(q->mid,qnew->mid)==0) {
			imsg("Ignoring duplicate entry in feed file %s\n",q->mid);
			/* I'm almost certain we should free qnew here */
			return;
		}
	}
	if (last==NULL) {
		l->que = qnew;
	} else {
		last->next = qnew;
	}
	dmsg("link_add_msg %s\n",qnew->mid);
	qnew->next = NULL;
	qnew->state = L_CHECK;
	link_send_check(l,qnew);
	l->nque++;
}
void que_store(Que *q, Feed *feed)
{
	FILE *f;
	char *body;
	int got=0;
	int total=0;
	body = que_get(q, 0, &got,&total,feed);
	if (body==NULL) return;
	f = fopen(file_work("dfeed.tout"),"wb");
	if (f==NULL) return;
	fwrite(body,1,total,f);
	fclose(f);
}

static int ntout;
void link_que_trim(Flink *l)
{
	Que *q;
	int doclose=FALSE;
	time_t t = time(NULL);

	for (q=l->que;q!=NULL;q=q->next) {
		if (t-q->tread > 600) {
			ntout++;
			emsg("Que entry timed out %s %d seconds\n",q->mid,t - q->tread);
			que_store(q,l->feed);
			q = link_que_remove(l,q->mid); if (q==NULL) return ;
			l->feed->nrej++;
			if (l->cur==q) {
				emsg("Timeout on sending entry, will close link\n");
				doclose = TRUE;
			}
			pending_release(q,l);	/* Record that item is no longer needed for this feed */
			if (doclose) {
				link_close(l,"Timeout sending que entry took over 10 minutes");
			}
			return; /* Can't do more now */
		}
	}
}
void link_que_fillx(Flink *l);
void link_que_fill(Flink *l)
{
	stats_start(S_QUE_FILL);
	link_que_fillx(l);
	stats_end(S_QUE_FILL);	
}
void link_que_fillx(Flink *l)
{
	Que *q;
	int n=0;
	if (l->state!=F_SENDING) {
		imsg("Link_que_fill %s:%d not in sending state yet\n",l->feed->site,l->id);
		return;
	}
	for (;!link_isfull(l);) {
		q = link_next(l);
		if (q==NULL) break;
		link_add_msg(l,q);
		n++;
	}
	dmsg("Link_que_fill %s:%d, que'd %d, %d now in queue\n",l->feed->site,l->id,n,l->nque);
}
void feed_write_upto(Feed *feed,char *mid)
{
	FILE *f;
	f = fopen(file_sendupto(feed->site),"w");
	if (f==NULL) return;
	fprintf(f,"%s\n",mid);
	fclose(f);
}
char *feed_read_upto(Feed *feed)
{
	FILE *f;
	char *s;
	static char bf[BFSZ];
	f = fopen(file_sendupto(feed->site),"r");
	if (f==NULL) return NULL;
	fgets(bf,200,f);
	fclose(f);
	s = strchr(bf,'\n');
	if (s!=NULL) *s = 0;
	return bf;
}
Que *link_next(Flink *l)
{
	int r;
	Que *q;
	int bkt,offset,len,spool,pile,ntries;
	char mid[BFSZ];
/*	if (l->feed->fscan==NULL) imsg("fscan is null for this feed at the moment\n");*/
	if (l->feed->fscan==NULL) return NULL;
again:
	r = send_read(l->feed->fscan,mid,&bkt,&offset,&len,&spool,&pile,&ntries);
	if (!r) {
		if (feof(l->feed->fscan)) {
			imsg("End of file for that send file %.100s\n",file_sendtmp(l->feed->site));
			fclose(l->feed->fscan); l->feed->fscan = NULL;
			remove(file_sendupto(l->feed->site));
			if (remove(file_sendtmp(l->feed->site))!=0) {
				emsg("Couldn't remove (%s) (%s) \n",file_sendtmp(l->feed->site),strerror(errno));
			}
			feeds_scan(l->feed);
		} else {
			imsg("Bad line in file, reading next line %d %.100s\n",ftell(l->feed->fscan),file_sendtmp(l->feed->site));
			goto again;
		}
		return NULL;
	}
	l->feed->nscan++;
	q = malloc(sizeof(Que));
	dmsg("Malloc que %p %s\n",q,mid);
	memset(q,0,sizeof(Que));
	q->mid = strdup(mid);
	q->bkt = bkt;
	q->offset = offset;
	q->len = len;
	q->spool = spool;
	q->pile = pile;
	q->tread = time(NULL);
	q->ntries = ntries;
	return q;
}
void que_free(Que *q)
{
	if (q->mid!=NULL) free(q->mid);
	q->mid = NULL;
	dmsg("Free q=%p\n",q);
	free(q);
}
char *file_bucket(Que *q)
{
	static char bf[BFSZ];
	if (q->pile==0) {
	  sprintf(bf,"%s/db_%d.itm",file_spool(q->spool),q->bkt);
	} else {
	  sprintf(bf,"%s/db_%d_%d.itm",file_spool(q->spool),q->bkt,q->pile);
	}
	return bf;
}
char *show_ten(char *s)
{
	static char bf[BFSZ];
	ncpy(bf,s,20);
	return bf;
}
char *xque_get(Que *q, int from, int *got, int *total, Feed *feed);
char *que_get(Que *q, int from, int *got, int *total, Feed *feed)
{
	char *s;
	stats_start(S_XREAD);	
	s = xque_get(q,from,got,total,feed);
	stats_end(S_XREAD);	
	return s;
}
char *xque_get(Que *q, int from, int *got, int *total, Feed *feed)
{
	int r;
	char bf[BFSZ];
	static char *data;
	static int datalen;
	int at_1,at_2,diff;
	static char lastmid[BFSZ];
	FILE *f;
	int len = q->len - HEAD_LEN - 1;
	
	if (datalen==0) {
		datalen = atoi(ini_get(I_ITEM_MAX));
		if (datalen<1000000) datalen = 1000000;
		datalen += 10000;
		data = malloc(datalen);
		g_data = data;
	}
	if (len+1000>datalen) {
		emsg("GETFAIL Item is too large to send %d>%d, increase item_max in dnews.conf\n",len+1000,datalen);
		return NULL;
	}

	if (strcmp(lastmid,q->mid)!=0) {	
		stats_start(S_OPEN);
		f = bfile_open(q);
		stats_end(S_OPEN);
		if (f==NULL) {
			imsg("GETFAIL Bucket file deleted %s %s\n",file_bucket(q),strerror(errno));
			goto failed;
		}
		stats_start(S_FSEEK);

		at_1 = ftell(f);
		at_2 = q->offset+HEAD_LEN-1;
		diff = at_2 - at_1;
		if ((diff>0) && (diff<BFSZ)) {
		 if ((int)fread(bf,1,diff,f)!=diff) {
			emsg("GETFAIL Could not read to item bkt=%d offset=%d %s\n",q->bkt,q->offset+HEAD_LEN-1,strerror(errno));
			goto failed;			 
		 }
		} else {
		 if (fseek(f,at_2,SEEK_SET)!=0) {
			emsg("GETFAIL Could not seek to item bkt=%d offset=%d %s\n",q->bkt,q->offset+HEAD_LEN-1,strerror(errno));
			goto failed;
		 }
		}	
		stats_end(S_FSEEK);
		stats_start(S_FREAD);	
		r = fread(data,1,len+2,f);
		stats_end(S_FREAD);	
		data[len+2] = 0;
		if (r!=len+2) {emsg("GETFAIL Couldn't read entire item %d/%d %s\n",r,len+2,strerror(errno)); goto failed;}
		if (data[0]!=':') {emsg("GETFAIL Data did not start with a colon (%s)\n",show_ten(data)); goto failed;}
		if (data[len+1]!='^') {emsg("GETFAIL Data did not end with a hat (%s)\n",show_ten(data - 20)); goto failed;}
		if (data[len]!='\n') {emsg("GETFAIL Data did not end with a lf (%s)\n",show_ten(data - 20)); goto failed;}
		data[len+1] = 0;
	} 

	if (data[len]!='\n') {emsg("GETFAIL2 Data did not end with a lf, should never happen (%s)\n",show_ten(data - 20)); goto failed;}
	
	ncpy(lastmid,q->mid,200);
	dmsg("que_get %s\n",q->mid);
	*total = len;
	*got = len - from; /* r - from;*/
	return data + from + 1;
failed:
	return NULL;
}
#define MAXBFILE 10
static FILE *bfile[MAXBFILE];
static int bfile_bkt[MAXBFILE];
FILE *bfile_open(Que *q)
{
	int i;
	static int next;
	for (i=0; i<MAXBFILE; i++) {
		if (bfile_bkt[i]==q->bkt) break;
	}
	if (i>=MAXBFILE) {
		next++;
		if (next>=MAXBFILE) next = 0;
		if (bfile[next]!=NULL) fclose(bfile[next]);
		bfile[next] = NULL;
		bfile_bkt[next] = q->bkt;
		i = next;
	}
	if (bfile[i]!=NULL) return bfile[i];
	bfile[i] = fopen(file_bucket(q),"rb");
	if (bfile[i]!=NULL) setvbuf(bfile[i],NULL,_IOFBF,32767);
	dmsg("Open [%d] %s\n",i,file_bucket(q));
	return bfile[i];
}
char *pad_left(char *s, int n)
{
	static char bf[BFSZ];
	int add = n - strlen(s);
	if (add<1) {strcpy(bf,s); return bf;}
	strcpy(bf,"                   ");
	strcpy(bf+add,s);
	return bf;
}

static int start_time;
void stats_reset(void)
{
	start_time = time(NULL);
}
void feeds_status(FILE *fout)
{
	int i;
	Feed *f;
	int days;
	Que *q;
	int iscomp;
	char fsize[BFSZ];
	static int up_time;
	int fsz,fsz2;
	int total_time;
	double tsentk,tsentn,comp,raw;
	Flink *l;
	int qtotal;
	
	if (start_time==0) start_time = time(NULL)-1;
	if (up_time==0) up_time = time(NULL)-1;
	total_time = time(NULL) - start_time;

	tsentk = 0;
	tsentn = 0;

	fprintf(fout,"   Items      Dup      Rej     Lost    File   Sitename\n");
	fprintf(fout,"   Bytes      K/s  Items/s  Scanned    OK/s   InQue\n\n");
	for (i=0; i<nfeeds; i++) {
		f = &feeds[i];
		fsz = lib_filesize(file_feed(f->site));
		fsz2 = lib_filesize(file_send(f->site));
		fsz2 += lib_filesize(file_sendtmp(f->site));
		if (fsz2>1000000) {
			sprintf(fsize,"%d/%dM",fsz2/1000000,fsz/1000000);
		} else {
			sprintf(fsize,"%d/%dk",fsz2/1000,fsz/1000);
		}
		strcpy(fsize,pad_left(fsize,7));
		qtotal = 0;
		iscomp = FALSE;
		raw = comp = 0;
		for (l=f->links;l!=NULL;l=l->next) {
			qtotal += l->nque;
			if (l->chan!=NULL) if (l->chan->iscompress) {
				raw += l->chan->rawbytes;
				comp += l->chan->compressedbytes;
				iscomp = TRUE;
			}
		}

		fprintf(fout,"%s%8d %8d %8d %8d %s   %s\n%8.0fmb %6.0fk/s %5.1f/s %6d %5.1f/s   %d\n---------->%s\n---------->%s\n\n"
			,f->isudp ? "UDP " : "",f->nok,f->ndup,f->nrej,f->ndel,fsize,f->site
			,f->nbytes/1000000.0,f->nbytes/1000.0/total_time
			,(double) f->nok/total_time
			,f->nscan,(double) (f->nok+f->ndup+f->nrej+f->ndel)/total_time
			,qtotal
			,f->reason,f->problem);
		if (iscomp) {
			if (raw==0) raw = 1;
			fprintf(fout,"  Compression %3.2f%% of original size\n",100.0*comp/raw);
		}

/*
should add something like this, and write to real dnews.log? 

  > Feed out summary %s Posts=%d Rejects=%d Dup=%d Queued=%d Requed=%d
File=%dM Posts=%.2f/s Total=%.2f/s Total=%.0fk
*/
		tsentk += f->nbytes/1024;
		tsentn += f->nok;
		if (ini_true(I_DFEED_DEBUG)) {
			for (l=f->links;l!=NULL;l=l->next) {
				fprintf(fout,"-debug- link [%d] %d\n",l->id,l->nque);
				for (q=l->que; q!=NULL; q=q->next) {
				fprintf(fout,"  -debug- que %d %s\n",q->state,q->mid);
					
				}
			}
		}
	}
	fprintf(fout,"Number scanned from que files %d\n",nscan);
	fprintf(fout,"Total sent %.0fmb %.0fkbytes/s, Total items sent ok %.0fk %3.1f/s\n",
		tsentk/1024,tsentk/total_time,tsentn/1000,tsentn/total_time); 
	if (nnul>0) {
		imsg("Historical Nulls found %d %s\n",nnul,nulmsg);
		fprintf(fout,"Historical Nulls %d, %s\n",nnul,nulmsg);
	}
	if (ntout>0) {
		/*imsg("Historical timeouts on que entries %d\n",ntout);*/
		fprintf(fout,"Historical timeouts on que entries %d\n",ntout);
	}
	days = (time(NULL) - up_time)/(60*60*24);
	fprintf(fout,"Uptime %d days %2.2f hours\n"
		,days
		,(double) (time(NULL) - up_time)/(60*60) - days*24);

	
}
void que_show(Flink *l)
{
	Que *q;
	for (q=l->que; q!=NULL; q=q->next) {
		dmsg(" -que entry %d %s\n",q->state,q->mid);
	}
}

