/* Copyright (C) 1995 NetWin World, New Zealand, All rights reserved. */
/* Memory routines which allow compaction   */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include "mem.h"
#define INDNEWS  TRUE
#ifdef INDNEWS
#include "emsg.h"
#include "lib.h"
#else
#define emsg printf
#define imsg printf
#define dmsg printf
#define FALSE 0
#define TRUE (!FALSE)
#endif

static struct BLOCK *fblock;
static int islocked;

/* Use the raw malloc and free.*/

static int pstr_total,pstr_used,pstr_peak;

#ifndef INDNEWS
main()
{
        int i,j;
        char bf[100];
        int sz;
        char *s;
        char **p[1000];
        int *m,**mm;
        for (i=0; i<100; i++) {
                sz = ((123456*i) % 100000) + 10;
                p[i] = p_malloc(sz);    
                for (s = *p[i]; s < (*p[i] + sz); s++) *s = i;

                /*p[200] = p_malloc(20); */    /* To cause fragmentation */     
                printf("Allocated %p \n",p[i]);
                if (p[i]==NULL) printf("Returned null\n");
                if (i>10) {
                        printf("call free %d %p %p \n",i-10,p[i-10],*(p[i-10]));
                        p_free(p[i-10]); p[i-10] = NULL;
                }

                for (j=0; j<100; j++) {
                        if (p[j]!=NULL) {
                                sz = ((123456*j) % 100000) + 10;
                                for (s = *p[j]; s < (*p[j] + sz); s++) {
                                        if (*s != j) printf("+++++Datalost\n");
                                }
                        }
                }
                printf("I=%d\n",i);
        }
        p_show();
        for (i=0; i<100; i++) if (p[i]!=NULL) p_free(p[i]);
        p_show();
        printf("Status %s\n",p_status());
        return 0;
}
#endif

void p_lock(char **x)
{
        islocked = TRUE;
}
void p_unlock(char **x)
{
        islocked = FALSE;
}
char *p_show_blockn(int n)
{
        Block *b;
        Piece *p;
        int total=0;
	int i;
	int np=0;
	static char bf[BFSZ];

        for (b=fblock,i=0; (b!=NULL) && (i<n); b=b->next) i++;
	if (b==NULL) return "";
        for (p = b->start; p!=NULL; p = p->next) {
		if (!p->isfree) {total += p->size; np++;}
        }
	sprintf(bf,"  Block [%d] Used %8dk of %8dk in %d pieces, free %d %d \015\012",n,total/1000,b->size/1000,np,b->totfree,b->size-total);
	return bf;
}

void p_show(void)
{
        Block *b;
        Piece *p;
        int total=0;

        for (b=fblock; b!=NULL; b=b->next) {
                imsg("==Block size %d biggest %d upto %p \n",b->size,b->biggest,b->upto);
                total += b->size;
                for (p = b->start; p!=NULL; p = p->next) {
                        imsg("Piece %p size %d free %d last %p\n",p,p->size,p->isfree,p->last);
                }
        }
        imsg("Total pmalloced %d \n",total);
}
void p_show_block(Block *b)
{
        Piece *p;
        int total=0;

        imsg("Block size %d biggest %d upto %p \n",b->size,b->biggest,b->upto);
        total += b->size;
        for (p = b->start; p!=NULL; p = p->next) {
                imsg("Piece %p size %d free %d last %p\n",p,p->size,p->isfree,p->last);
        }
}

char **p_malloc(int sz)
{
        Block *b,*xb;
        int newsz;
        static Block *lastb;
        Piece *p;
	static int done_init=FALSE;
        if (sz==0) {emsg("Error: malloc zero bytes \n"); return NULL;}
        sz += 3;
        sz = sz & 0x5ffffffc;
        dmsg("mem: pmalloc:  %d \n",sz);                
        pstr_used += sz;
        if (pstr_used>pstr_peak) pstr_peak = pstr_used;
        if (fblock==NULL) {fblock = p_init(10000); lastb = fblock;}

        for (b=fblock; b!=NULL; b=b->next) {
                xb = b;
		if (b->totfree<50000) continue; /* Dont try and squeeze the last bit in*/ 
       		if (b->totfree<sz) continue; /* This one won't fit*/ 
                if (p_try1(b,&p,sz)) goto gotit;
                p_compress1(b);
                if (p_try1(b,&p,sz)) goto gotit;
                p_compress2(b);
                if (p_try1(b,&p,sz)) goto gotit;
        }



/*        b = lastb;
        if (p_try1(b,&p,sz)) goto gotit;
        for (b=fblock; b!=NULL; b=b->next) {
                if (p_try1(b,&p,sz)) goto gotit;
        }

        dmsg("mem: Simple Compress\n");
        for (b=fblock; b!=NULL; b=b->next) {
                p_compress1(b);
                if (p_try1(b,&p,sz)) goto gotit;
        }
        dmsg("mem: Real Compress\n");
        for (b=fblock; b!=NULL; b=b->next) {
                xb = b;
                p_compress2(b);
                if (p_try1(b,&p,sz)) goto gotit;
        }
  */
        newsz = sz + 10000;
        if (newsz < xb->size) newsz = xb->size + 20000; 
     	done_init = TRUE;
	if (newsz<sz) newsz = sz + 10000;
        xb->next = p_init(newsz); b = xb->next;
        /* p_show(); */
        if (p_try1(b,&p,sz)) goto gotit;
        emsg("p_malloc failed\n"); exit(1);
gotit:
        lastb = b;
        return (char **) p;
}

void p_flush(void)
{
        Block *b,*last,*nextb;
	Piece *p,*nextp;
	int total,np;

	last = NULL;
        dmsg("mem: compressing strings\n");
        if (fblock==NULL) return;
        for (b=fblock; b!=NULL; b=nextb) {
dmsg("mem: compressing strings1\n");
                p_compress1(b);
dmsg("mem: compressing strings2\n");
                p_compress2(b);
		nextb = b->next;
#ifdef WIN32
		total = 0;  np = 0;
	        for (p = b->start; p!=NULL; p = p->next) {
			if (!p->isfree) {total += p->size; np++;}
	        }

		nextb = b->next;
		if ( (b->totfree == b->size-total) && (b!=fblock) && (np==0) ) {
		 	emsg("Freeing string block that is empty %p size %d\n",b,b->totfree);
		        pstr_total -= b->totfree;

		        for (p = b->start; p!=NULL; p = nextp) {
				nextp = p->next;
				myfree(p);
		        }
			if (last==NULL) emsg("Fault in block logic\n");
			else last->next = b->next;
			myfree(b->base);
			myfree(b);
		} else {
			last = b;
		}
#endif
        }
        dmsg("mem: compressing strings done\n");
}

Block *p_init(int sz)
{
        Block *b;
        dmsg("mem: p_init allocate %d bytes \n",sz);
        pstr_total += sz;
        b = mymalloc(sizeof(Block));
        if (b==NULL) {emsg("malloc failed in pinit\n"); exit(3); }
	b->totfree = sz;
        b->size = sz;
        b->base = mymalloc(sz);
        if (b->base==NULL) {emsg("malloc failed in pinit base %d\n",sz); exit(3); }
        b->start = p_new(sz,b->base,b,NULL,NULL);
        b->upto = b->start;
        b->biggest = sz;
        b->next = NULL;
        return b;
}
int p_try1(Block *b, Piece **p, int sz) /* Try and find a block that is big enough */
{
        if (b->biggest<sz) return FALSE;
        *p = b->upto;
        if ((*p)->isfree) if ((*p)->size>=sz) return p_got(b,*p,sz);
        for (*p = b->start; (*p)!=NULL; (*p) = (*p)->next) {
                if ((*p)->isfree) if ((*p)->size>=sz) return p_got(b,*p,sz);    
        }
        return FALSE;
}
int p_got(Block *b, Piece *p, int sz)
{
        p->next = p_split(p, sz);
        p->size = sz;
        p->islocked = FALSE;
        p->isfree = FALSE;
        b->upto = p->next;
        b->biggest = p->next->size; /* Not entirely true but safe enough*/
	b->totfree -= sz;
        return TRUE;
}
Piece *p_new(int sz, void *data, Block *b, Piece *last, Piece *next)
{
        Piece *n = mymalloc(sizeof(Piece));
        n->size = sz;
        n->block = b;
        n->magic = 12345;
        n->isfree = TRUE;
        n->data = data;
        n->last = last;
        n->next = next;
        if (next != NULL) {
                next->last = n;
        }
        return n;
}
Piece *p_split(Piece *p, int sz)
{
        Piece *n = p_new(p->size-sz,p->data+sz,p->block,p,p->next);
        return n;
}
void p_free(char **pp)
{
        Piece *p = (Piece *) pp;

        dmsg("mem: pfree %p size %d \n",p,p->size);             
        if (p->magic!=12345) {emsg("Panic in p_free, magic  = %d \n",p->magic); return;}
        if (p->isfree) {emsg("Panic in p_free\n"); return;}
        pstr_used -= p->size;
	p->block->totfree += p->size;

        p->isfree = TRUE;
        for (;p!=NULL;) {
                if (! p->isfree) break;
                if (p->next == NULL) break;
                if (p->next->isfree) p_unsplit(p,"freedelete");
                else break;
        }
        p = (Piece *) pp;
/*      if (p->block->biggest< p->size) p->block->biggest = p->size;*/
}

void p_unsplit(Piece *p, char *text)
{
        Piece *old = p->next;
/*      dmsg("mem: Unsplit (%s) %p - %p   %d %d \n",text,p,p->next,p->isfree,p->next->isfree);*/
        p->size += p->next->size;       
        if (p->block->biggest< p->size) p->block->biggest = p->size;
        p->next = p->next->next;
        if (p->next != NULL) p->next->last = p;
        myfree(old);
        p->block->upto = p;
}
int p_compress1(Block *b)
{
        Piece *p;

        /* Remove duplicate empty cells */      
        for (p = b->start; p!=NULL; p = p->next) {
again:          if (p->isfree) if (p->next!=NULL) if (p->next->isfree) {
                        p_unsplit(p,"compress1");
                        goto again;
                }
                if (p->isfree) if (b->biggest < p->size)  b->biggest = p->size;
        }
        b->upto = b->start;
        return TRUE;
}
/* Move pieces between blocks to make more space */
int p_compress3(Block *dst, Block *src)
{
	return 0;	
}

/* Move blocks around to make more space */
int p_compress2(Block *b)
{
        Piece *p,*dst,*pnext;
	
        dmsg("Compress block %p \n",b);
        dmsg("Compress block %p size %d \n",b,b->size);
        if (islocked) {emsg("mem: compress2 can't due to lock\n"); return FALSE;}
        /* Search for free one, then search for next used one and move it */    
        for (p = b->start; p!=NULL; p = p->next) {
                if (p->isfree) {
                        dst = p;
                        for (p=p->next; p!=NULL; p = pnext) {
				pnext = p->next;
                                if (p->isfree) p_unsplit(dst,"compress2"); else break;
                        }
                        if (p!=NULL) { /* Then move this real block to DST*/
                                dmsg("p_compress2 Move %p \n",p);
                                memmove(dst->data,p->data,p->size);
                                p->data = dst->data;
                                dst->data = p->data+p->size;
                                p->last = dst->last;
                                if (dst->last ==NULL) b->start = p;
                                else dst->last->next = p;
                                if (p->next!=NULL) p->next->last = dst;
                                dst->next = p->next;
                                p->next = dst;
                                dst->last = p;
                                dmsg("p_compress2 Move  done %p \n",p);
                        }
                }
                if (p==NULL) break;
        }
        b->upto = b->start;
        return TRUE;
}

/* Just an idea, not used yet. */
void *raw_malloc(int x)
{
        return malloc(x);
}
void raw_free(void *x)
{
        free(x);
}
char *p_status(void)
{
        static char bf[100];
        sprintf(bf,"%dk/%dk/%dk",pstr_used/1000,pstr_peak/1000,pstr_total/1000);
        return bf;
}
